"""__init__.py"""
__version__ = "0.2.5"


class GoogleTagManagerScopes:
    """Scopes available in the Google Tag Manager API."""

    MANAGE_USERS = "https://www.googleapis.com/auth/tagmanager.manage.users"
    READONLY = "https://www.googleapis.com/auth/tagmanager.readonly"
    EDIT_CONTAINERS = "https://www.googleapis.com/auth/tagmanager.edit.containers"
    EDIT_CONTAINERVERSIONS = (
        "https://www.googleapis.com/auth/tagmanager.edit.containerversions"
    )
    PUBLISH = "https://www.googleapis.com/auth/tagmanager.publish"
    MANAGE_ACCOUNTS = "https://www.googleapis.com/auth/tagmanager.manage.accounts"
    DELETE_CONTAINERS = "https://www.googleapis.com/auth/tagmanager.delete.containers"


AUTH_SCOPES = [
    GoogleTagManagerScopes.MANAGE_USERS,
    GoogleTagManagerScopes.READONLY,
    GoogleTagManagerScopes.EDIT_CONTAINERS,
    GoogleTagManagerScopes.EDIT_CONTAINERVERSIONS,
    GoogleTagManagerScopes.PUBLISH,
    GoogleTagManagerScopes.MANAGE_ACCOUNTS,
    GoogleTagManagerScopes.DELETE_CONTAINERS,
]

SERVICE_NAME = "tagmanager"
SERVICE_VERSION = "v2"
CREDENTIALS_FILE_NAME = "_auth_credentials.json"
CLIENT_SECRET_FILE = "client_secret.json"
HEADLESS_AUTH = False

GENERIC_REQUEST_ERROR = "Returned an error response for your request."
NO_LIVE_VERSION_ERROR = "Published container version not found"
BACKEND_ERROR = "Backend Error"
RATE_LIMIT_ERROR = "Quota exceeded for quota group"

DEFAULT_HTTP_TIMEOUT_SEC = 60
RATE_LIMIT_CALLS = 24
RATE_LIMIT_PERIOD = 100



"""manager.py"""
import logging
from googleapiclient import errors

import gtm_manager.account


class GTMManager(gtm_manager.base.GTMBase):
    """Authenticates a users base gtm access.
    """

    def __init__(self, **kwargs):

        super().__init__(**kwargs)

        self.accounts_service = self.service.accounts()  # pylint: disable=E1101

    def list_accounts(self):
        """Loads from the API and lists all GTM Accounts that a user has access to.

        Returns:
            A list of :class:`gtm_manager.account.GTMAccount` that the user has access to.
        """
        try:
            request = self.accounts_service.list()
            response = request.execute()
            return [
                gtm_manager.account.GTMAccount(account=x, service=self.service)
                for x in response.get("account") or []
            ]
        except errors.HttpError as error:
            logging.error(error)
            return []



"""account.py"""
import gtm_manager.base
import gtm_manager.container
import gtm_manager.permission


class GTMAccount(gtm_manager.base.GTMBase):
    """Open a specific GTM account.

    Args:
        account (dict): An account API representation of the GTM Account. If provided, the account
            will be not be loaded from the API. `account` or `path` argument
            must be set.
        path (str): The GTM API path to the account, i.e "accounts/1234". If provided instead of 
            `account`, the account representation will be loaded from the API. `path` or `account`
            argument must be set.
        **kwargs: Additional keyword args to initialize the base class.
    """

    def __init__(self, account=None, path=None, **kwargs):

        super().__init__(**kwargs)

        self.accounts_service = self.service.accounts()  # pylint: disable=E1101

        if account:
            pass
        elif path:
            account = self._get_account(path)
        else:
            raise ValueError("Please pass either an account obj or account path.")

        self._name = account.get("name")
        self._shareData = account.get("shareData", False)
        self._tagManagerUrl = account.get("tagManagerUrl")
        self._fingerprint = account.get("fingerprint")
        self._path = account.get("path")
        self._accountId = account.get("accountId")

        self._list_containers = None
        self._list_permissions = None

        self._raw_body = account

    @property
    def name(self):
        """str: Account display name
        """
        return self._name

    @property
    def shareData(self):
        """bool: Whether the account shares data anonymously with Google and others
        """
        return self._shareData

    @property
    def tagManagerUrl(self):
        """str: Auto generated link to the tag manager UI
        """
        return self._tagManagerUrl

    @property
    def fingerprint(self):
        """str: The fingerprint of the GTM Account as computed at
        storage time. This value is recomputed whenever the account is modified.
        """
        return self._fingerprint

    @property
    def path(self):
        """str: GTM Account's API relative path
        """
        return self._path

    @property
    def raw_body(self):
        """obj: The raw asset body as returend from the API
        """
        return self._raw_body

    @property
    def accountId(self):
        """str: The Account ID uniquely identifies the GTM Account
        """
        return self._accountId

    def list_containers(self, refresh=True):
        """Load from API and list all containers in this account.

        Args:
            refresh (bool): If containers have already been loaded in the API, force another API
                request to get the latest list of containers.

        Returns:
            A list of :class:`gtm_manager.permission.GTMContainer`.
        """
        if self._list_containers is None or refresh:
            self._refresh_list_containers()

        return self._list_containers

    def list_permissions(self, refresh=True):
        """Load from API and list all permissions in this account.

        Args:
            refresh (bool): If permissions have already been loaded in the API, force another API
                request to get the latest list of permissions.

        Returns:
            A list of :class:`gtm_manager.permission.GTMPermission`.
        """
        if self._list_permissions is None or refresh:
            self._refresh_list_permissions()

        return self._list_permissions

    def update(self, name, shareData):
        """Update the current accounts `name` and `shareData` parameters via the API.

        Args:
            name (str): Account display name
            shareData (bool): Share data setting
        """
        asset_body = {"name": name, "shareData": shareData}

        request = self.accounts_service.update(path=self.path, body=asset_body)
        response = request.execute()
        self.__init__(account=response, service=self.service)

    def create_container(self, name, usage_context="web", domain_name=None, notes=None):
        """Create a new container in the current account.

        Args:
            name (str): Container display name
            usage_context (str): Usage Contexts for the Container.
                Valid values include: web, android, or ios.
            domain_name (str): List of domain names associated with the Container.
            notes (str): Container notes
        """
        asset_body = {
            "name": name,
            "usageContext": [usage_context] if usage_context else ["web"],
        }

        if domain_name:
            asset_body["domainName"] = [domain_name]
        if notes:
            asset_body["notes"] = notes

        request = self.accounts_service.containers().create(
            parent=self.path, body=asset_body
        )
        response = request.execute()
        return gtm_manager.container.GTMContainer(
            container=response, service=self.service
        )

    def _get_account(self, path):
        """_get_container"""
        request = self.accounts_service.get(path=path)
        response = request.execute()
        return response

    def _refresh_list_containers(self):
        """_refresh_list_container"""
        request = self.accounts_service.containers().list(parent=self.path)
        response = request.execute()
        self._list_containers = [
            gtm_manager.container.GTMContainer(container=x, service=self.service)
            for x in response.get("container") or []
        ]

    def _refresh_list_permissions(self):
        """_refresh_list_permissions"""
        request = self.accounts_service.user_permissions().list(parent=self.path)
        response = request.execute()
        self._list_permissions = [
            gtm_manager.permission.GTMPermission(permission=x, service=self.service)
            for x in response.get("userPermission") or []
        ]



"""permission.py"""

import gtm_manager


class GTMPermission(gtm_manager.base.GTMBase):
    """Open a specific GTM Permission.

    Args:
        permission (dict): An API representation of the GTM Permission. If provided, the resource
            will be not be loaded from the API. :code:`permission` or :code:`path` argument must be
            set.
        path (str): The API path to the resource, i.e.
            "accounts/1234/permissions/1234". If provided instead of :code:`permission`,
            the representation will be loaded from the API. :code:`path` or :code:`permission`
            argument must be set.
        **kwargs: Additional keyword args to initialize the base class.
    """

    def __init__(self, permission=None, path=None, **kwargs):
        super().__init__(**kwargs)

        self.permission_service = (
            self.service.accounts().user_permissions()
        )  # pylint: disable=E1101

        if permission:
            pass
        elif path:
            permission = self._get_permission(path)
        else:
            raise ValueError("Please pass either a permission obj or permission path.")

        self._containerAccess = [
            GTMContainerAccess(x) for x in permission.get("containerAccess") or []
        ]
        self._path = permission.get("path")
        self._accountAccess = permission.get("accountAccess")
        self._emailAddress = permission.get("emailAddress")
        self._accountId = permission.get("accountId")

    def __repr__(self):
        return "{} {}".format(self.path, self._emailAddress)

    @property
    def path(self):
        """str: GTM Permission's API relative path."""
        return self._path

    @property
    def containerAccess(self):
        """str: GTM Permission's API relative path."""
        return self._containerAccess

    @property
    def accountAccess(self):
        """str: GTM Permission's API relative path."""
        return self._accountAccess

    @property
    def emailAddress(self):
        """str: GTM Permission's API relative path."""
        return self._emailAddress

    @property
    def accountId(self):
        """str: GTM Permission's API relative path."""
        return self._accountId

    def _get_permission(self, path):
        """_get_container"""
        request = self.permission_service.get(path=path)
        response = request.execute()
        return response


class GTMContainerAccess(object):
    """GTMContainerAccess"""

    def __init__(self, containerAccess):
        self.containerId = containerAccess.get("containerId")
        self.permission = containerAccess.get("permission")



"""base.py"""
from googleapiclient import discovery

from gtm_manager.utils_auth import build_http

from gtm_manager import SERVICE_NAME, SERVICE_VERSION


class GTMBase(object):
    """GTMBase"""

    def __init__(self, service=None):
        if service:
            self.service = service
        else:
            self.service = discovery.build(
                SERVICE_NAME, SERVICE_VERSION, http=build_http(), cache_discovery=False
            )



"""tag.py"""
import copy

import gtm_manager.base
import gtm_manager.parameter
from gtm_manager.utils import param_dict


class GTMTag(gtm_manager.base.GTMBase):
    """Open a specific GTM Tag.

    Args:
        tag (dict): An API representation of the GTM Tag. If provided, the resource will be not be
            loaded from the API. :code:`tag` or :code:`path` argument must be set.
        path (str): The API path to the resource, i.e.
            "accounts/1234/containers/1234/workspaces/1234/tags/123". If provided instead of
            :code:`tag`, the representation will be loaded from the API. :code:`path` or :code:`tag`
            argument must be set.
        parent (str): Required, when the instance is initialized with a :code:`tag` argument to
            explizitly set the parent path, i.e. "accounts/1234/containers/1234/workspaces/1234"
        **kwargs: Additional keyword args to initialize the base class.
    """

    def __init__(self, tag=None, path=None, parent=None, **kwargs):
        super().__init__(**kwargs)

        self.tags_service = (
            self.service.accounts().containers().workspaces().tags()
        )  # pylint: disable=E1101

        if tag:
            pass
        elif path:
            tag = self._get_tag(path)
        else:
            raise ValueError(
                "Please pass either a container obj and parent or container path."
            )

        self._paused = tag.get("paused")
        self._setupTag = tag.get("setupTag")
        self._firingRuleId = tag.get("firingRuleId", [])
        self._accountId = tag.get("accountId")
        self._teardownTag = tag.get("teardownTag")
        self._priority = tag.get("priority")
        self._workspaceId = tag.get("workspaceId")
        self._parameter = tag.get("parameter", [])
        self._parentFolderId = tag.get("parentFolderId")
        self._scheduleStartMs = tag.get("scheduleStartMs")
        self._scheduleEndMs = tag.get("scheduleEndMs")
        self._containerId = tag.get("containerId")
        self._tagFiringOption = tag.get("tagFiringOption")
        self._tagId = tag.get("tagId")
        self._blockingRuleId = tag.get("blockingRuleId", [])
        self._tagManagerUrl = tag.get("tagManagerUrl")
        self._fingerprint = tag.get("fingerprint")
        self._firingTriggerId = tag.get("firingTriggerId", [])
        self._name = tag.get("name")
        self._type = tag.get("type")
        self._notes = tag.get("notes")
        self._liveOnly = tag.get("liveOnly")
        self._blockingTriggerId = tag.get("blockingTriggerId", [])
        self._path = path or "{}/tags/{}".format(parent, self._tagId)

        self._parameter = [
            gtm_manager.parameter.GTMParameter(x) for x in self._parameter
        ]

    @property
    def paused(self):
        """bool: Indicates whether the tag is paused, which prevents the tag from firing.
        """
        return self._paused

    @property
    def setupTag(self):
        """list: The list of setup tags. Currently we only allow one.
        """
        return self._setupTag

    @property
    def firingRuleId(self):
        """bool: Firing rule IDs. A tag will fire when any of the listed rules are true and all of
        its blockingRuleIds (if any specified) are false.
        """
        return self._firingRuleId

    @property
    def accountId(self):
        """str: GTM Account ID.
        """
        return self._accountId

    @property
    def teardownTag(self):
        """list: The list of teardown tags. Currently we only allow one.
        """
        return self._teardownTag

    @property
    def priority(self):
        """obj: Represents a Google Tag Manager Parameter. # User defined numeric priority of the
        tag. Tags are fired asynchronously in order of priority. Tags with higher numeric value fire
        first. A tag's priority can be a positive or negative value. The default value is 0.
        """
        return self._priority

    @property
    def workspaceId(self):
        """str: GTM Workspace ID.
        """
        return self._workspaceId

    @property
    def parameter(self):
        """list: The tag's parameters.
        """
        return self._parameter

    @property
    def parentFolderId(self):
        """str: Parent folder id.
        """
        return self._parentFolderId

    @property
    def scheduleStartMs(self):
        """str: The start timestamp in milliseconds to schedule a tag.
        """
        return self._scheduleStartMs

    @property
    def scheduleEndMs(self):
        """str: The end timestamp in milliseconds to schedule a tag.
        """
        return self._scheduleEndMs

    @property
    def containerId(self):
        """str: GTM Container ID.
        """
        return self._containerId

    @property
    def tagFiringOption(self):
        """str: Option to fire this tag.
        """
        return self._tagFiringOption

    @property
    def tagId(self):
        """str: The Tag ID uniquely identifies the GTM Tag.
        """
        return self._tagId

    @property
    def blockingRuleId(self):
        """list: Blocking rule IDs. If any of the listed rules evaluate to true, the tag will not
        fire.
        """
        return self._blockingRuleId

    @property
    def tagManagerUrl(self):
        """str: Auto generated link to the tag manager UI
        """
        return self._tagManagerUrl

    @property
    def fingerprint(self):
        """str: The fingerprint of the GTM Tag as computed at storage time. This value is recomputed
        whenever the tag is modified.
        """
        return self._fingerprint

    @property
    def path(self):
        """str: GTM Tag's API relative path.
        """
        return self._path

    @property
    def firingTriggerId(self):
        """list: Firing trigger IDs. A tag will fire when any of the listed triggers are true and
        all of its blockingTriggerIds (if any specified) are false.
        """
        return self._firingTriggerId

    @property
    def name(self):
        """str: Tag display name.
        """
        return self._name

    @property
    def type(self):
        """str: GTM Tag Type.
        """
        return self._type

    @property
    def notes(self):
        """str: User notes on how to apply this tag in the container.
        """
        return self._notes

    @property
    def liveOnly(self):
        """bool: If set to true, this tag will only fire in the live environment (e.g. not in
        preview or debug mode).
        """
        return self._liveOnly

    @property
    def blockingTriggerId(self):
        """list: Blocking trigger IDs. If any of the listed triggers evaluate to true, the tag will
        not fire.
        """
        return self._blockingTriggerId

    @property
    def parameter_dict(self):
        """dict: Deepcopy of GTMParameters acceable via their key value.
        """
        return param_dict(copy.deepcopy(self._parameter))

    def update(self, refresh=False, parameter=None, **kwargs):
        """Update the current tag. The GTM API does not support a partial update. Therfore, this
        method will send all fields expliztily set in the method arguments and those cached in the 
        instance properties.

        GTMParameters passed in a list as the `parameter` argument, will be merged recursivly with
        the exsisting parameters based on their parameter key.

        All other API resource properties can be overwritten by specifying the property name as
        keyword arguments on the method call. 

        Args:
            refresh (bool): Force a refresh of the entire GTMTag instance to prevent implicitly
                sending stale property data.
            parameter (list): :class:`gtm_manager.parameter.GTMParameter` s to be merged recursivly
                with the exsisting parameters based on their parameter key.
            **kwargs: Additional resource properties to update with this call.

        Raises:
            ValueError
        """
        if parameter and not isinstance(parameter, list):
            raise ValueError(
                "'parameter' has to be a list of :class:`GTMParameters` or 'None'."
            )

        if refresh:
            self.__init__(path=self._path, service=self.service)

        default_asset = {
            "paused": self._paused,
            "setupTag": self._setupTag,
            "firingRuleId": self._firingRuleId,
            "teardownTag": self._teardownTag,
            "priority": self._priority,
            "parentFolderId": self._parentFolderId,
            "scheduleStartMs": self._scheduleStartMs,
            "scheduleEndMs": self._scheduleEndMs,
            "tagFiringOption": self._tagFiringOption,
            "blockingRuleId": self._blockingRuleId,
            "firingTriggerId": self._firingTriggerId,
            "name": self._name,
            "type": self._type,
            "notes": self._notes,
            "liveOnly": self._liveOnly,
            "blockingTriggerId": self._blockingTriggerId,
        }
        update_asset = {**default_asset, **kwargs}

        if parameter:
            parameter_dict = {**param_dict(self._parameter), **param_dict(parameter)}
            parameter = list(parameter_dict.values())
        else:
            parameter = self._parameter

        update_asset["parameter"] = [x.to_obj() for x in parameter]

        update_asset = {k: v for k, v in update_asset.items() if v is not None}

        request = self.tags_service.update(path=self.path, body=update_asset)
        response = request.execute()
        self.__init__(tag=response, service=self.service)

    def _get_tag(self, path):
        """_get_tag"""
        request = self.tags_service.get(path=path)
        response = request.execute()
        return response

    def delete(self):
        """Delete the current tag.
        """
        request = self.tags_service.delete(path=self._path)
        request.execute()



"""utils_auth.py"""
import json

from httplib2 import Http

from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from google_auth_httplib2 import AuthorizedHttp
from googleapiclient.errors import HttpError

from ratelimit import limits, sleep_and_retry, RateLimitException
from retrying import retry

from gtm_manager.exceptions import AuthError
from gtm_manager import (
    DEFAULT_HTTP_TIMEOUT_SEC,
    RATE_LIMIT_CALLS,
    RATE_LIMIT_PERIOD,
    BACKEND_ERROR,
    RATE_LIMIT_ERROR,
)


class LimittedHttp(Http):
    """LimittedHttp"""

    @retry(
        stop_max_attempt_number=6,
        wait_exponential_multiplier=1000,
        wait_exponential_max=10000,
        retry_on_exception=lambda x: BACKEND_ERROR in str(x),
    )
    @retry(
        stop_max_attempt_number=6,
        wait_exponential_multiplier=10000,
        wait_exponential_max=100000,
        retry_on_exception=lambda x: RATE_LIMIT_ERROR in str(x),
    )
    @sleep_and_retry
    @limits(calls=RATE_LIMIT_CALLS, period=RATE_LIMIT_PERIOD)
    def request(self, *args, **kwargs):
        """request"""
        try:
            return super().request(*args, **kwargs)
        except HttpError as error:
            if RATE_LIMIT_ERROR in str(error):
                raise RateLimitException("too many calls", RATE_LIMIT_PERIOD)
            raise error


def get_credentials():
    """get_credentials"""
    from . import AUTH_SCOPES, CLIENT_SECRET_FILE, HEADLESS_AUTH, CREDENTIALS_FILE_NAME

    try:
        credentials = Credentials.from_authorized_user_file(CREDENTIALS_FILE_NAME)
    except FileNotFoundError:
        credentials = None
    except ValueError:
        credentials = None

    if credentials:
        return credentials

    try:
        flow = InstalledAppFlow.from_client_secrets_file(
            client_secrets_file=CLIENT_SECRET_FILE, scopes=AUTH_SCOPES
        )
    except FileNotFoundError:
        raise AuthError(
            "Please provide a client secrete file as %r" % CLIENT_SECRET_FILE
        )

    if not HEADLESS_AUTH:
        credentials = flow.run_local_server()
    else:
        credentials = flow.run_console()

    creds_data = {
        "token": None,
        "refresh_token": credentials.refresh_token,
        "token_uri": credentials.token_uri,
        "client_id": credentials.client_id,
        "client_secret": credentials.client_secret,
        "scopes": credentials.scopes,
    }

    with open(CREDENTIALS_FILE_NAME, "w") as outfile:
        json.dump(creds_data, outfile)

    return credentials


def build_http():
    """build_http"""
    return AuthorizedHttp(
        get_credentials(), http=LimittedHttp(timeout=DEFAULT_HTTP_TIMEOUT_SEC)
    )



"""variable.py"""
import copy

import gtm_manager.base
import gtm_manager.parameter
from gtm_manager.utils import param_dict


class GTMVariable(gtm_manager.base.GTMBase):
    """Open a specific GTM Variable.

    Args:
        trigger (dict): An API representation of the GTM trigger. If provided, the resource will be
            not be loaded from the API. :code:`trigger` or :code:`path` argument must be set.
        path (str): The API path to the resource, i.e.
            "accounts/1234/containers/1234/workspaces/1234/trigger/123". If provided instead of
            :code:`trigger`, the representation will be loaded from the API. :code:`path` or
            :code:`trigger` argument must be set.
        parent (str): Required, when the instance is initialized with a trigger argument to
            explizitly set the parent path, i.e. "accounts/1234/containers/1234/workspaces/1234"
        **kwargs: Additional keyword args to initialize the base class.
    """

    def __init__(self, variable=None, path=None, parent=None, **kwargs):
        super().__init__(**kwargs)

        self.variables_service = (
            self.service.accounts().containers().workspaces().variables()
        )  # pylint: disable=E1101

        if variable:
            pass
        elif path:
            variable = self._get_variable(path)
        else:
            raise ValueError("Please pass either a container obj or container path.")

        self._scheduleStartMs = variable.get("scheduleStartMs")
        self._scheduleEndMs = variable.get("scheduleEndMs")
        self._name = variable.get("name")
        self._variableId = variable.get("variableId")
        self._type = variable.get("type")
        self._notes = variable.get("notes")
        self._enablingTriggerId = variable.get("enablingTriggerId")
        self._workspaceId = variable.get("workspaceId")
        self._tagManagerUrl = variable.get("tagManagerUrl")
        self._fingerprint = variable.get("fingerprint")
        self._accountId = variable.get("accountId")
        self._parameter = variable.get("parameter") or []
        self._parentFolderId = variable.get("parentFolderId")
        self._disablingTriggerId = variable.get("disablingTriggerId")
        self._containerId = variable.get("containerId")

        self._path = path or "{}/variables/{}".format(parent, self._variableId)

        self._parameter = [
            gtm_manager.parameter.GTMParameter(x) for x in self._parameter
        ]

    @property
    def scheduleStartMs(self):
        """str: The start timestamp in milliseconds to schedule a variable.
        """
        return self._scheduleStartMs

    @property
    def scheduleEndMs(self):
        """str: The end timestamp in milliseconds to schedule a variable.
        """
        return self._scheduleEndMs

    @property
    def name(self):
        """str: Variable display name.
        """
        return self._name

    @property
    def variableId(self):
        """str: The Variable ID uniquely identifies the GTM Variable.
        """
        return self._variableId

    @property
    def type(self):
        """str: GTM Variable Type.
        """
        return self._type

    @property
    def notes(self):
        """str: User notes on how to apply this variable in the container.
        """
        return self._notes

    @property
    def enablingTriggerId(self):
        """list: For mobile containers only - A list of trigger IDs for enabling
        conditional variables; the variable is enabled if one of the enabling
        triggers is true while all the disabling triggers are false.
        Treated as an unordered set.
        """
        return self._enablingTriggerId

    @property
    def workspaceId(self):
        """str: GTM Workspace ID.
        """
        return self._workspaceId

    @property
    def tagManagerUrl(self):
        """str: Auto generated link to the tag manager UI
        """
        return self._tagManagerUrl

    @property
    def fingerprint(self):
        """str: The fingerprint of the GTM Variable as computed at storage
        time. This value is recomputed whenever the variable is modified.
        """
        return self._fingerprint

    @property
    def path(self):
        """str: GTM Variable's API relative path.
        """
        return self._path

    @property
    def accountId(self):
        """str: GTM Account ID.
        """
        return self._accountId

    @property
    def parameter(self):
        """list: The variable's parameters.
        """
        return self._parameter

    @property
    def parentFolderId(self):
        """str: Parent folder id.
        """
        return self._parentFolderId

    @property
    def disablingTriggerId(self):
        """list: For mobile containers only - A list of trigger IDs for disabling
        conditional variables; the variable is enabled if one of the enabling
        trigger is true while all the disabling trigger are false. Treated as an
        unordered set.
        """
        return self._disablingTriggerId

    @property
    def containerId(self):
        """str: GTM Container ID.
        """
        return self._containerId

    @property
    def parameter_dict(self):
        """dict: GTM parameters acceable via their key value.
        """
        return param_dict(copy.deepcopy(self._parameter))

    def _get_variable(self, path):
        """_get_variable"""
        request = self.variables_service.get(path=path)
        response = request.execute()
        return response

    def constant_value(self):
        """constant_value"""
        if self.type == "c":
            return self.parameter[0]["value"]
        else:
            return None

    def update(self, refresh=False, parameter=None, **kwargs):
        """Update the current variable. The GTM API does not support a partial update. Therfore,
        this method will send all fields expliztily set in the method arguments and those cached in
        the instance properties.

        GTMParameters passed in a list as the :code:`parameter` argument, will be merged recursivly
        with the exsisting parameters based on their parameter key.

        All other API resource properties can be overwritten by specifying the property name as
        keyword arguments on the method call.

        Args:
            refresh (bool): Force a refresh of the entire GTMVariable instance to prevent implicitly
                sending stale property data.
            parameter (list): :class:`gtm_manager.parameter.GTMParameter` list to be merged
                recursivly with the exsisting parameters based on their parameter key.
            **kwargs: Additional resource properties to update with this call.

        Raises:
            ValueError
        """
        if refresh:
            self.__init__(path=self._path, service=self.service)

        default_asset = {
            "scheduleStartMs": self._scheduleStartMs,
            "scheduleEndMs": self._scheduleEndMs,
            "name": self._name,
            "variableId": self._variableId,
            "type": self._type,
            "notes": self._notes,
            "enablingTriggerId": self._enablingTriggerId,
            "workspaceId": self._workspaceId,
            "tagManagerUrl": self._tagManagerUrl,
            "fingerprint": self._fingerprint,
            "path": self._path,
            "accountId": self._accountId,
            "parentFolderId": self._parentFolderId,
            "disablingTriggerId": self._disablingTriggerId,
            "containerId": self._containerId,
        }
        update_asset = {**default_asset, **kwargs}

        if parameter:
            parameter_dict = {**param_dict(self._parameter), **param_dict(parameter)}
            parameter = list(parameter_dict.values())
        else:
            parameter = self._parameter

        update_asset["parameter"] = [x.to_obj() for x in parameter]

        update_asset = {k: v for k, v in update_asset.items() if v is not None}

        request = self.variables_service.update(path=self.path, body=update_asset)
        response = request.execute()
        self.__init__(variable=response, service=self.service)

    def delete(self):
        """Delete the current variable.
        """
        request = self.variables_service.delete(path=self._path)
        request.execute()



"""built_in_variable.py"""


class GTMBuiltInVariable(object):
    """Open a specifc GTM Built In Variable

    Args:
        - built_in_variable (dict): An API representation of the GTM Built In Variable.
    """

    def __init__(self, built_in_variable):
        self._path = built_in_variable.get("path")
        self._accountId = built_in_variable.get("accountId")
        self._containerId = built_in_variable.get("containerId")
        self._workspaceId = built_in_variable.get("workspaceId")
        self._type = built_in_variable.get("type")
        self._name = built_in_variable.get("name")

    @property
    def path(self):
        """str: Account display name
        """
        return self._path

    @property
    def accountId(self):
        """str: Account display name
        """
        return self._accountId

    @property
    def containerId(self):
        """str: Account display name
        """
        return self._containerId

    @property
    def workspaceId(self):
        """str: Account display name
        """
        return self._workspaceId

    @property
    def name(self):
        """str: Account display name
        """
        return self._name

    @property
    def type(self):
        """str: Account display name
        """
        return self._type



"""gtm workspace.py"""
import gtm_manager.base
import gtm_manager.version
import gtm_manager.tag
import gtm_manager.trigger
import gtm_manager.variable
import gtm_manager.folder
import gtm_manager.built_in_variable
from gtm_manager.exceptions import TagNotFound, TriggerNotFound, VariableNotFound


class GTMWorkspace(gtm_manager.base.GTMBase):
    """Open a specific GTM Workspace.

    Args:
        workspace (dict): An API representation of the GTM Workspace. If provided, the
            resource will be not be loaded from the API. `workspace` or `path` argument must be
            set.
        path (str): The API path to the resource, i.e
            "accounts/1234/containers/1234/workspaces/1234". If provided instead of `workspace`,
            the representation will be loaded from the API. `path` or `workspace` argument must be
            set.
        **kwargs: Additional keyword args to initialize the base class.
    """

    def __init__(self, workspace=None, path=None, **kwargs):
        super().__init__(**kwargs)

        self.workspaces_service = self.service.accounts().containers().workspaces()

        if workspace:
            pass
        elif path:
            workspace = self._get_workspace(path)
        else:
            raise ValueError(
                "Please pass either a workspace API representation or workspace path."
            )

        self._description = workspace.get("description")
        self._name = workspace.get("name")
        self._workspaceId = workspace.get("workspaceId")
        self._tagManagerUrl = workspace.get("tagManagerUrl")
        self._fingerprint = workspace.get("fingerprint")
        self._path = workspace.get("path")
        self._accountId = workspace.get("accountId")
        self._containerId = workspace.get("containerId")
        self._tags = []
        self._triggers = []
        self._variables = []
        self._folders = []
        self._built_in_variables = []
        self._quick_preview = None

    def __repr__(self):
        return "<GTM Workspace: {}>".format(self.name)

    @property
    def description(self):
        """str: Account display name
        """
        return self._description

    @property
    def name(self):
        """str: Account display name
        """
        return self._name

    @property
    def workspaceId(self):
        """str: Account display name
        """
        return self._workspaceId

    @property
    def tagManagerUrl(self):
        """str: Account display name
        """
        return self._tagManagerUrl

    @property
    def fingerprint(self):
        """str: Account display name
        """
        return self._fingerprint

    @property
    def path(self):
        """str: Account display name
        """
        return self._path

    @property
    def accountId(self):
        """str: Account display name
        """
        return self._accountId

    @property
    def containerId(self):
        """str: Account display name
        """
        return self._containerId

    def quick_preview(self, refresh=True):
        """Get a quick_preview of the current workspace state from the API.

        Args:
            Args:
            refresh (bool): If quick_preview has already been loaded from the API, force another API
                request to get the latest quick_preview.

        Returns:
            An instance of :class:`gtm_manager.version.GTMVersion`
        """
        if refresh or not self._quick_preview:
            self._refresh_quick_preview()

        return self._quick_preview

    def delete(self):
        """Delete the current workspace.
        """
        request = self.workspaces_service.delete(path=self.path)
        request.execute()

    def clear_all_assets(self, refresh=True):
        """Clear all assets from the current workspace.

        Args:
            refresh (bool): If quick_preview has already been loaded from the API, force another API
                request to get the latest quick_preview.

        """
        if refresh or not self._quick_preview:
            self._refresh_quick_preview()

        if self._quick_preview.tag:
            for tag in self._quick_preview.tag:
                tag.delete()

        if self._quick_preview.trigger:
            for trigger in self._quick_preview.trigger:
                trigger.delete()

        for _ in range(10):
            if self._quick_preview.variable:
                for variable in reversed(self._quick_preview.variable):
                    variable.delete()
                self._refresh_quick_preview()
            else:
                break

        if self._quick_preview.folder:
            for folder in self._quick_preview.folder:
                folder.delete()

    def trigger_map(self, refresh=True):
        """Create a trigger map from the current workspace.

        Args:
            refresh (bool): If quick_preview has already been loaded from the API, force another API
                request to get the latest quick_preview.

        Returns:
            A dict having all triggerNames and triggerIds as strings keys with their corresponding
            triggerIds and triggerNames as string values. The *All Pages* trigger will always be
            added as default. For example::

                {
                    "2147479553": "All Pages",
                    "All Pages": "2147479553",
                    "Click - My Button": "1",
                    "1": "Click - My Button",
                }
        """
        if refresh or not self._quick_preview:
            self._refresh_quick_preview()

        trigger_map = {"2147479553": "All Pages", "All Pages": "2147479553"}

        if not self._quick_preview.trigger:
            return trigger_map

        for trigger in self._quick_preview.trigger:
            trigger_map[trigger.name] = trigger.triggerId
            trigger_map[trigger.triggerId] = trigger.name

        return trigger_map

    def folder_map(self, refresh=True):
        """Create a folder map from the current workspace.

        Args:
            refresh (bool): If quick_preview has already been loaded from the API, force another API
                request to get the latest quick_preview.

        Returns:
            A dict having all folderNames and folderIds as strings keys with their corresponding
            folderIds and folderName as string values. For example::

                {
                    "My Folder": "1",
                    "My Other Folder": "2",
                    "1":"My Folder",
                    "2": "My Other Folder",
                }
        """
        if refresh or not self._quick_preview:
            self._refresh_quick_preview()

        folder_map = {}

        if not self._quick_preview.folder:
            return folder_map

        for folder in self._quick_preview.folder:
            folder_map[folder.name] = folder.folderId
            folder_map[folder.folderId] = folder.name

        return folder_map

    def disable_built_ins(self, built_in_type):
        """Disable built-ins in the current workspace.

        Possible values are documented `here <https://developers.google.com/tag-manager/api/v2/reference/accounts/containers/workspaces/built_in_variables/create>`

        Args:
            built_in_type (list of str): A list of the built in types as strings the call should
                disable.
        """
        self.workspaces_service.built_in_variables().delete(
            path=self.path, type=built_in_type
        )

    def create_tag(self, asset_body):
        """Create a tag in the current workspace.

        @TODO: Add writeable fields.

        Args:
            asset_body (dict): An API representation of a GTM Tag.

        Returns:
            An instance of :class:`gtm_manager.tag.GTMTag`
        """
        request = self.workspaces_service.tags().create(
            parent=self.path, body=asset_body
        )
        return gtm_manager.tag.GTMTag(
            tag=request.execute(), parent=self.path, service=self.service
        )

    def create_trigger(self, asset_body):
        """Create a trigger in the current workspace.

        @TODO: Add writeable fields.

        Args:
            asset_body (dict): An API representation of a GTM Trigger.

        Returns:
            An instance of :class:`gtm_manager.trigger.GTMTrigger`
        """
        request = self.workspaces_service.triggers().create(
            parent=self.path, body=asset_body
        )
        return gtm_manager.trigger.GTMTrigger(
            trigger=request.execute(), parent=self.path, service=self.service
        )

    def create_variable(self, asset_body):
        """Create a variable in the current workspace.

        @TODO: Add writeable fields.

        Args:
            asset_body (dict): An API representation of a GTM Variable.

        Returns:
            An instance of :class:`gtm_manager.variable.GTMVariable`
        """
        request = self.workspaces_service.variables().create(
            parent=self.path, body=asset_body
        )
        return gtm_manager.variable.GTMVariable(
            variable=request.execute(), parent=self.path, service=self.service
        )

    def create_folder(self, name, notes=""):
        """Create a folder in the current workspace.

        @TODO: Add writeable fields.

        Args:
            asset_body (dict): An API representation of a GTM Folder.

        Returns:
            An instance of :class:`gtm_manager.folder.GTMFolder`
        """
        request = self.workspaces_service.folders().create(
            parent=self.path, body={"name": name, "notes": notes}
        )
        return gtm_manager.folder.GTMFolder(
            folder=request.execute(), parent=self.path, service=self.service
        )

    def create_build_ins(self, asset_body):
        """Create built-ins in the current workspace.

        Args:
            built_in_type (list of str): A list of the built in types as strings the call should
                enable.
        """
        request = self.workspaces_service.built_in_variables().create(
            parent=self.path, type=asset_body
        )
        response = request.execute()
        return [
            gtm_manager.built_in_variable.GTMBuiltInVariable(x)
            for x in response.get("builtInVariable") or []
        ]

    def list_tags(self, refresh=False):
        """List all tags from the current workspace.

        Args:
            refresh (bool): If quick_preview has already been loaded from the API, force another API
                request to get the latest quick_preview.

        Returns:
            A list of :class:`gtm_manager.tag.GTMTag`.
        """
        if self._tags is None or refresh:
            request = self.workspaces_service.tags().list(parent=self.path)

            response = request.execute()
            self._tags = [
                gtm_manager.tag.GTMTag(tag=x, service=self.service, parent=self.path)
                for x in response.get("tag") or []
            ]
        return self._tags

    def list_triggers(self, refresh=False):
        """List all triggers from the current workspace.

        Args:
            refresh (bool): If quick_preview has already been loaded from the API, force another API
                request to get the latest quick_preview.

        Returns:
            A list of :class:`gtm_manager.triggers.GTMTrigger`.
        """
        if self._triggers is None or refresh:
            request = self.workspaces_service.triggers().list(parent=self.path)

            response = request.execute()
            self._triggers = [
                gtm_manager.trigger.GTMTrigger(
                    trigger=x, service=self.service, parent=self.path
                )
                for x in response.get("trigger") or []
            ]
        return self._triggers

    def list_variables(self, refresh=False):
        """List all variables from the current workspace.

        Args:
            refresh (bool): If quick_preview has already been loaded from the API, force another API
                request to get the latest quick_preview.

        Returns:
            A list of :class:`gtm_manager.variable.GTMVariable`.
        """
        if self._variables is None or refresh:
            request = self.workspaces_service.variables().list(parent=self.path)

            response = request.execute()
            self._variables = [
                gtm_manager.variable.GTMVariable(
                    variable=x, service=self.service, parent=self.path
                )
                for x in response.get("variable") or []
            ]
        return self._variables

    def list_folders(self, refresh=False):
        """List all folders from the current workspace.

        Args:
            refresh (bool): If quick_preview has already been loaded from the API, force another API
                request to get the latest quick_preview.

        Returns:
            A list of :class:`gtm_manager.folder.GTMFolder`.
        """

        if self._folders is None or refresh:
            request = self.workspaces_service.folders().list(parent=self.path)

            response = request.execute()
            self._folders = [
                gtm_manager.folder.GTMFolder(
                    folder=x, service=self.service, parent=self.path
                )
                for x in response.get("folder") or []
            ]
        return self._folders

    def list_built_in_variables(self, refresh=False):
        """List all built in variables from the current workspace.

        Args:
            refresh (bool): If quick_preview has already been loaded from the API, force another API
                request to get the latest quick_preview.

        Returns:
            A list of :class:`gtm_manager.built_in_variable.GTMBuiltInVariable`.
        """

        if self._built_in_variables is None or refresh:
            request = self.workspaces_service.built_in_variables().list(
                parent=self.path
            )

            response = request.execute()
            self._built_in_variables = [
                gtm_manager.built_in_variable.GTMBuiltInVariable(built_in_variable=x)
                for x in response.get("builtInVariable") or []
            ]
        return self._built_in_variables

    def get_tag_by_name(self, tag_name, refresh=False):
        """Get a GTM Tag from the GTM Workspace by its name.

        Args:
            tag_name (str): The exact name to look for in the GTM Workspace.
            refresh (bool): If quick_preview has already been loaded from the API, force another API
                request to get the latest quick_preview.

        Returns:
            An instance of :class:`gtm_manager.tag.GTMTag`.

        Raises:
            :class:`gtm_manager.exceptions.TagNotFound`
        """
        tags = self.list_tags(refresh=refresh)

        for tag in tags:
            if tag.name == tag_name:
                return tag

        raise TagNotFound(tag_name, self.path)

    def get_trigger_by_name(self, trigger_name, refresh=False):
        """Get a GTM Trigger from the GTM Workspace by its name.

        Args:
            trigger_name (str): The exact name to look for in the GTM Workspace.
            refresh (bool): If quick_preview has already been loaded from the API, force another API
                request to get the latest quick_preview.

        Returns:
            An instance of :class:`gtm_manager.trigger.GTMTrigger`.

        Raises:
            :class:`gtm_manager.exceptions.TriggerNotFound`
        """
        triggers = self.list_triggers(refresh=refresh)

        for trigger in triggers:
            if trigger.name == trigger_name:
                return trigger

        raise TriggerNotFound(trigger_name, self.path)

    def get_variable_by_name(self, variable_name, refresh=False):
        """Get a GTM Variable from the GTM Workspace by its name.

        Args:
            variable_name (str): The exact name to look for in the GTM Workspace.
            refresh (bool): If quick_preview has already been loaded from the API, force another API
                request to get the latest quick_preview.

        Returns:
            An instance of :class:`gtm_manager.variable.GTMVariable`.

        Raises:
            :class:`gtm_manager.exceptions.VariableNotFound`
        """
        variables = self.list_variables(refresh=refresh)

        for variable in variables:
            if variable.name == variable_name:
                return variable

        raise VariableNotFound(variable_name, self.path)

    def create_version(self, name, notes=""):
        """Create a new version from the current state of the workspace.

        @TODO: add descroption for sync or publishing errors

        Args:
            name (str): The GTM Version display name
            notes (str); The GTM Version notes

        Returns:
            An instance of :class:`gtm_manager.version.GTMVersion`.
        """
        request = self.workspaces_service.create_version(
            path=self.path, body={"name": name, "notes": notes}
        )
        response = request.execute()

        version = gtm_manager.version.GTMVersion(
            version=response.get("containerVersion"), service=self.service
        )

        if version.containerVersionId == "0":
            raise Exception(
                "Could not create version from workspace. Please visit {} for details".format(
                    self.tagManagerUrl
                )
            )

        return version

    def _get_workspace(self, path):
        """_get_workspace"""
        request = self.workspaces_service.get(path=path)
        response = request.execute()
        return response

    def _refresh_quick_preview(self):
        """_refresh_quick_preview"""
        request = self.workspaces_service.quick_preview(path=self.path)
        response = request.execute()
        self._quick_preview = gtm_manager.version.GTMVersion(
            version=response.get("containerVersion"),
            workspaceId=self.workspaceId,
            service=self.service,
        )
        return self._quick_preview

    def get_status(self):
        """Get the status with all workspace changes.

        @TODO: Add examples

        Returns:
            A list with all workspace changes.
        """
        request = self.workspaces_service.getStatus(path=self.path)
        return request.execute()

    def sync(self):
        """Synchronize the workspace to the latest GTM Version.

        Returns:
            @TODO: What is returned?
        """
        request = self.workspaces_service.sync(path=self.path)
        return request.execute()



"""trigger.py"""
import gtm_manager.base
import gtm_manager.parameter
from gtm_manager.utils import param_dict


class GTMTrigger(gtm_manager.base.GTMBase):
    """Open a specific GTM Trigger.

    Args:
        trigger (dict): An API representation of the GTM Trigger. If provided, the resource will be
            not be loaded from the API. :code:`trigger` or :code:`path` argument must be set.
        path (str): The API path to the resource, i.e.
            "accounts/1234/containers/1234/workspaces/1234/trigger/123". If provided instead of
            :code:`trigger`, the representation will be loaded from the API. :code:`path` or
            :code:`trigger` argument must be set.
        parent (str): Required, when the instance is initialized with a :code:`trigger` argument to
            explizitly set the parent path, i.e. "accounts/1234/containers/1234/workspaces/1234"
        **kwargs: Additional keyword args to initialize the base class.
    """

    def __init__(self, trigger=None, path=None, parent=None, **kwargs):
        super().__init__(**kwargs)

        self.triggers_service = (
            self.service.accounts().containers().workspaces().triggers()
        )  # pylint: disable=E1101

        if trigger:
            pass
        elif path:
            trigger = self._get_trigger(path)
        else:
            raise ValueError("Please pass either a container obj or container path.")

        self._maxTimerLengthSeconds = trigger.get("maxTimerLengthSeconds")
        self._totalTimeMinMilliseconds = trigger.get("totalTimeMinMilliseconds")
        self._uniqueTriggerId = trigger.get("uniqueTriggerId")
        self._verticalScrollPercentageList = trigger.get("verticalScrollPercentageList")
        self._horizontalScrollPercentageList = trigger.get(
            "horizontalScrollPercentageList"
        )
        self._containerId = trigger.get("containerId")
        self._waitForTagsTimeout = trigger.get("waitForTagsTimeout")
        self._accountId = trigger.get("accountId")
        self._waitForTags = trigger.get("waitForTags")
        self._intervalSeconds = trigger.get("intervalSeconds")
        self._eventName = trigger.get("eventName")
        self._visibilitySelector = trigger.get("visibilitySelector")
        self._workspaceId = trigger.get("workspaceId")
        self._customEventFilter = trigger.get("customEventFilter")
        self._parameter = trigger.get("parameter") or []
        self._parentFolderId = trigger.get("parentFolderId")
        self._continuousTimeMinMilliseconds = trigger.get(
            "continuousTimeMinMilliseconds"
        )
        self._selector = trigger.get("selector")
        self._triggerId = trigger.get("triggerId")
        self._tagManagerUrl = trigger.get("tagManagerUrl")
        self._fingerprint = trigger.get("fingerprint")
        self._visiblePercentageMax = trigger.get("visiblePercentageMax")
        self._name = trigger.get("name")
        self._visiblePercentageMin = trigger.get("visiblePercentageMin")
        self._type = trigger.get("type")
        self._notes = trigger.get("notes")
        self._interval = trigger.get("interval")
        self._filter = trigger.get("filter")
        self._autoEventFilter = trigger.get("autoEventFilter")
        self._limit = trigger.get("limit")
        self._checkValidation = trigger.get("checkValidation")

        self._path = path or "{}/triggers/{}".format(parent, self._triggerId)

        self._parameter = [
            gtm_manager.parameter.GTMParameter(x) for x in self._parameter
        ]

    @property
    def maxTimerLengthSeconds(self):
        """obj: Represents a Google Tag Manager Parameter. - Max time to fire Timer Events (in
        seconds). Only valid for AMP Timer trigger.
        """
        return self._maxTimerLengthSeconds

    @property
    def totalTimeMinMilliseconds(self):
        """obj:  Represents a Google Tag Manager Parameter. - A visibility trigger minimum total
        visible time (in milliseconds). Only valid for AMP Visibility trigger.
        """
        return self._totalTimeMinMilliseconds

    @property
    def uniqueTriggerId(self):
        """obj: Represents a Google Tag Manager Parameter. - Globally unique id of the trigger that
        auto-generates this (a Form Submit, Link Click or Timer listener) if any. Used to make
        incompatible auto-events work together with trigger filtering based on trigger ids. This
        value is populated during output generation since the tags implied by triggers don"t exist
        until then. Only valid for Form Submit, Link Click and Timer triggers.
        """
        return self._uniqueTriggerId

    @property
    def verticalScrollPercentageList(self):
        """obj: Represents a Google Tag Manager Parameter. - List of integer percentage values for
        scroll triggers. The trigger will fire when each percentage is reached when the view is
        scrolled vertically. Only valid for AMP scroll triggers.
        """
        return self._verticalScrollPercentageList

    @property
    def horizontalScrollPercentageList(self):
        """obj: Represents a Google Tag Manager Parameter. - List of integer percentage values for
        scroll triggers. The trigger will fire when each percentage is reached when the view is
        scrolled horizontally. Only valid for AMP scroll triggers.
        """
        return self._horizontalScrollPercentageList

    @property
    def containerId(self):
        """str: GTM Container ID.
        """
        return self._containerId

    @property
    def waitForTagsTimeout(self):
        """obj: Represents a Google Tag Manager Parameter. - How long to wait (in milliseconds) for
        tags to fire when "waits_for_tags" above evaluates to true. Only valid for Form Submission
        and Link Click triggers.
        """
        return self._waitForTagsTimeout

    @property
    def accountId(self):
        """str: GTM Account ID.
        """
        return self._accountId

    @property
    def waitForTags(self):
        """str: Represents a Google Tag Manager Parameter. - Whether or not we should delay the form
        submissions or link opening until all of the tags have fired (by preventing the default
        action and later simulating the default action). Only valid for Form Submission and Link
        Click triggers.
        """
        return self._waitForTags

    @property
    def intervalSeconds(self):
        """obj: Represents a Google Tag Manager Parameter. - Time between Timer Events to fire (in
        seconds). Only valid for AMP Timer trigger.
        """
        return self._intervalSeconds

    @property
    def eventName(self):
        """obj: Represents a Google Tag Manager Parameter. - Name of the GTM event that is fired.
        Only valid for Timer triggers.
        """
        return self._eventName

    @property
    def visibilitySelector(self):
        """obj: Represents a Google Tag Manager Parameter. - A visibility trigger CSS selector (i.e.
        "-id"). Only valid for AMP Visibility trigger.
        """
        return self._visibilitySelector

    @property
    def workspaceId(self):
        """str: GTM Workspace ID.
        """
        return self._workspaceId

    @property
    def customEventFilter(self):
        """list: Used in the case of custom event, which is fired iff all Conditions are true.
        """
        return self._customEventFilter

    @property
    def parameter(self):
        """list: Additional parameters.
        """
        return self._parameter

    @property
    def parentFolderId(self):
        """str: Parent folder id.
        """
        return self._parentFolderId

    @property
    def continuousTimeMinMilliseconds(self):
        """obj: Represents a Google Tag Manager Parameter. - A visibility trigger minimum continuous
        visible time (in milliseconds). Only valid for AMP Visibility trigger.
        """
        return self._continuousTimeMinMilliseconds

    @property
    def selector(self):
        """obj: Represents a Google Tag Manager Parameter. - A click trigger CSS selector (i.e. "a",
        "button" etc.). Only valid for AMP Click trigger.
        """
        return self._selector

    @property
    def triggerId(self):
        """str: The Trigger ID uniquely identifies the GTM Trigger.
        """
        return self._triggerId

    @property
    def tagManagerUrl(self):
        """str: Auto generated link to the tag manager UI
        """
        return self._tagManagerUrl

    @property
    def fingerprint(self):
        """str: The fingerprint of the GTM Trigger as computed at storage time. This value is
        recomputed whenever the trigger is modified.
        """
        return self._fingerprint

    @property
    def visiblePercentageMax(self):
        """obj: Represents a Google Tag Manager Parameter. - A visibility trigger maximum percent
        visibility. Only valid for AMP Visibility trigger.
        """
        return self._visiblePercentageMax

    @property
    def path(self):
        """str: GTM Trigger"s API relative path.
        """
        return self._path

    @property
    def name(self):
        """str: Trigger display name.
        """
        return self._name

    @property
    def visiblePercentageMin(self):
        """obj: Represents a Google Tag Manager Parameter. - A visibility trigger minimum percent
        visibility. Only valid for AMP Visibility trigger.
        """
        return self._visiblePercentageMin

    @property
    def type(self):
        """str: Defines the data layer event that causes this trigger.
        """
        return self._type

    @property
    def notes(self):
        """str: User notes on how to apply this trigger in the container.
        """
        return self._notes

    @property
    def interval(self):
        """obj: Represents a Google Tag Manager Parameter. - Time between triggering recurring Timer
        Events (in milliseconds). Only valid for Timer triggers.
        """
        return self._interval

    @property
    def filter(self):
        """list: The trigger will only fire iff all Conditions are true.
        """
        return self._filter

    @property
    def autoEventFilter(self):
        """list: Used in the case of auto event tracking.
        """
        return self._autoEventFilter

    @property
    def limit(self):
        """obj: Represents a Google Tag Manager Parameter. - Limit of the number of GTM events this
        Timer Trigger will fire. If no limit is set, we will continue to fire GTM events until the
        user leaves the page. Only valid for Timer triggers.
        """
        return self._limit

    @property
    def checkValidation(self):
        """obj: Represents a Google Tag Manager Parameter. - Whether or not we should only fire tags
         if the form submit or link click event is not cancelled by some other event handler (e.g.
         because of validation). Only valid for Form Submission and Link Click triggers.
        """
        return self._checkValidation

    def update(self, refresh=False, parameter=None, **kwargs):
        """Update the current trigger. The GTM API does not support a partial update. Therfore, this
        method will send all fields expliztily set in the method arguments and those cached in the 
        instance properties.

        GTMParameters passed in a list as the :code:`parameter` argument, will be merged recursivly
        with the exsisting parameters based on their parameter key.

        All other API resource properties can be overwritten by specifying the property name as
        keyword arguments on the method call. 

        Args:
            refresh (bool): Force a refresh of the entire GTMTrigger instance to prevent implicitly
                sending stale property data.
            parameter (list): :class:`gtm_manager.parameter.GTMParameter` list to be merged
                recursivly with the exsisting parameters based on their parameter key.
            **kwargs: Additional resource properties to update with this call.

        Raises:
            ValueError
        """
        if refresh:
            self.__init__(path=self._path, service=self.service)

        default_asset = {
            "maxTimerLengthSeconds": self._maxTimerLengthSeconds,
            "totalTimeMinMilliseconds": self._totalTimeMinMilliseconds,
            "uniqueTriggerId": self._uniqueTriggerId,
            "verticalScrollPercentageList": self._verticalScrollPercentageList,
            "horizontalScrollPercentageList": self._horizontalScrollPercentageList,
            "containerId": self._containerId,
            "waitForTagsTimeout": self._waitForTagsTimeout,
            "accountId": self._accountId,
            "waitForTags": self._waitForTags,
            "intervalSeconds": self._intervalSeconds,
            "eventName": self._eventName,
            "visibilitySelector": self._visibilitySelector,
            "workspaceId": self._workspaceId,
            "customEventFilter": self._customEventFilter,
            "parentFolderId": self._parentFolderId,
            "continuousTimeMinMilliseconds": self._continuousTimeMinMilliseconds,
            "selector": self._selector,
            "triggerId": self._triggerId,
            "tagManagerUrl": self._tagManagerUrl,
            "fingerprint": self._fingerprint,
            "visiblePercentageMax": self._visiblePercentageMax,
            "path": self._path,
            "name": self._name,
            "visiblePercentageMin": self._visiblePercentageMin,
            "type": self._type,
            "notes": self._notes,
            "interval": self._interval,
            "filter": self._filter,
            "autoEventFilter": self._autoEventFilter,
            "limit": self._limit,
            "checkValidation": self._checkValidation,
        }
        update_asset = {**default_asset, **kwargs}

        if parameter:
            parameter_dict = {**param_dict(self._parameter), **param_dict(parameter)}
            parameter = list(parameter_dict.values())
        else:
            parameter = self._parameter

        update_asset["parameter"] = [x.to_obj() for x in parameter]

        update_asset = {k: v for k, v in update_asset.items() if v is not None}

        request = self.triggers_service.update(path=self.path, body=update_asset)
        response = request.execute()
        self.__init__(trigger=response, service=self.service)

    def _get_trigger(self, path):
        """_get_trigger"""
        request = self.triggers_service.get(path=path)
        response = request.execute()
        return response

    def delete(self):
        """Delete the current trigger.
        """
        request = self.triggers_service.delete(path=self._path)
        request.execute()



"""parameter.py"""
import copy


class GTMParameter(object):
    """GTMParameter"""

    def __init__(self, parameter):
        self.map = parameter.get("map")
        self.list = parameter.get("list")
        self.value = parameter.get("value")
        self.key = parameter.get("key")
        self.type = parameter.get("type")

        if self.list:
            self.list = [GTMParameter(x) for x in self.list]

    def __repr__(self):
        return "<GTM Parameter: {}>".format(self.key)

    def to_obj(self):
        """to_obj"""
        if self.list:
            self.list = [x.to_obj() for x in self.list]
            # new_list = []
            # for item in self.list:
            #     if isinstance(item, GTMParameter):
            #         new_list.append(item.to_obj())
            #     else:
            #         new_list.append(item)
            # self.list = new_list
        return {k: v for k, v in self.__dict__.items() if v is not None}

    def copy(self):
        """copy"""
        return copy.deepcopy(self)



"""folder.py"""

import gtm_manager.base


class GTMFolder(gtm_manager.base.GTMBase):
    """Open a specific GTM Folder.

    Args:
        folder (dict): An API representation of the GTM Folder. If provided, the resource will be not be
            loaded from the API. :code:`folder` or :code:`path` argument must be set.
        path (str): The API path to the resource, i.e.
            "accounts/1234/containers/1234/workspaces/1234/folders/123". If provided instead of
            :code:`folder`, the representation will be loaded from the API. :code:`path` or
            :code:`folder` argument must be set.
        parent (str): Required, when the instance is initialized with a :code:`folder` argument to
            explizitly set the parent path, i.e. "accounts/1234/containers/1234/workspaces/1234"
        **kwargs: Additional keyword args to initialize the base class.
    """

    def __init__(self, folder=None, path=None, parent=None, **kwargs):
        super().__init__(**kwargs)

        self.folders_service = (
            self.service.accounts().containers().workspaces().folders()
        )  # pylint: disable=E1101

        if folder:
            pass
        elif path:
            folder = self._get_folder(path)
        else:
            raise ValueError("Please pass either a folder obj or folder path.")

        self._containerId = folder.get("containerId")
        self._notes = folder.get("notes", "")
        self._workspaceId = folder.get("workspaceId")
        self._tagManagerUrl = folder.get("tagManagerUrl")
        self._fingerprint = folder.get("fingerprint")
        self._folderId = folder.get("folderId")
        self._accountId = folder.get("accountId")
        self._name = folder.get("name")
        self._path = path or "{}/folders/{}".format(parent, self.folderId)

    @property
    def containerId(self):
        """str: The container that this version was taken from.
        """
        return self._containerId

    @property
    def notes(self):
        """str: The container that this version was taken from.
        """
        return self._notes

    @property
    def workspaceId(self):
        """str: The container that this version was taken from.
        """
        return self._workspaceId

    @property
    def tagManagerUrl(self):
        """str: The container that this version was taken from.
        """
        return self._tagManagerUrl

    @property
    def fingerprint(self):
        """str: The container that this version was taken from.
        """
        return self._fingerprint

    @property
    def folderId(self):
        """str: The container that this version was taken from.
        """
        return self._folderId

    @property
    def accountId(self):
        """str: The container that this version was taken from.
        """
        return self._accountId

    @property
    def name(self):
        """str: The container that this version was taken from.
        """
        return self._name

    @property
    def path(self):
        """str: The container that this version was taken from.
        """
        return self._path

    def _get_folder(self, path):
        """_get_container"""
        request = self.folders_service.get(path=path)
        response = request.execute()
        return response

    def delete(self):
        """Delete the current folder.
        """
        request = self.folders_service.delete(path=self._path)
        request.execute()



"""setup.py"""
import re
import os
import codecs
import textwrap
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

DEPENDENCIES = (
    "google-auth-oauthlib",
    "google-auth",
    "google-api-python-client",
    "retrying",
    "ratelimit",
)


def read(*parts):
    """read"""
    with codecs.open(os.path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    """find_version"""
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name="gtm-manager",
    version=find_version("gtm_manager", "__init__.py"),
    description="An object-oriented helper library wrapping the Tag Manager API Client Library for Python.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=["gtm_manager"],
    author="Trakken Web Services",
    author_email="christian@trakken.de",
    classifiers=textwrap.dedent(
        """
        Development Status :: 4 - Beta
        Programming Language :: Python :: 3.5
        Programming Language :: Python :: 3.6
        Programming Language :: Python :: 3.7
        License :: OSI Approved :: MIT License
    """
    )
    .strip()
    .splitlines(),
    keywords=["gtm", "google tag manager"],
    install_requires=DEPENDENCIES,
    extras_require={
        "tests": ["pytest"],
        "docs": ["sphinx >= 1.8.2", "sphinx_rtd_theme"],
    },
)



"""excpetions.py"""


class GTMManagerException(Exception):
    """GTMManagerException"""

    pass


class AuthError(GTMManagerException):
    pass


class VariableNotFound(GTMManagerException):
    """VariableNotFound"""

    def __init__(self, variable_name, parent):
        super(VariableNotFound, self).__init__()
        self.variable_name = variable_name
        self.parent = parent

    def __str__(self):
        return "Variable {!r} not found in parent {!r}".format(
            self.variable_name, self.parent
        )


class TagNotFound(GTMManagerException):
    """TagNotFound"""

    def __init__(self, tag_name, parent):
        super(TagNotFound, self).__init__()
        self.tag_name = tag_name
        self.parent = parent

    def __str__(self):
        return "Tag {!r} not found in parent {!r}".format(self.tag_name, self.parent)


class TriggerNotFound(GTMManagerException):
    """TriggerNotFound"""

    def __init__(self, trigger_name, parent):
        super(TriggerNotFound, self).__init__()
        self.trigger_name = trigger_name
        self.parent = parent

    def __str__(self):
        return "Trigger {!r} not found in parent {!r}".format(
            self.trigger_name, self.parent
        )


class RateLimitExceeded(GTMManagerException):
    """RateLimitExceeded"""

    pass



"""container.py"""
from googleapiclient.errors import HttpError

import gtm_manager.base
import gtm_manager.workspace
import gtm_manager.version

from gtm_manager import NO_LIVE_VERSION_ERROR


class GTMContainer(gtm_manager.base.GTMBase):
    """Open a specific GTM Container.

    Args:
        container (dict): An API representation of the GTM Container. If provided, the
            container will be not be loaded from the API. `container` or `path` argument must be
            set.
        path (str): The API path to the resource, i.e "accounts/1234/containers/1234". If
            provided instead of `container`, the api representation will be loaded from the API.
            `path` or `container` argument must be set.
        **kwargs: Additional keyword args to initialize the base class.
    """

    def __init__(self, container=None, path=None, **kwargs):

        super().__init__(**kwargs)

        self.containers_service = (
            self.service.accounts().containers()
        )  # pylint: disable=E1101

        if container:
            pass
        elif path:
            container = self._get_container(path)
        else:
            raise ValueError("Please pass either a container obj or container path.")

        self._publicId = container.get("publicId")
        self._containerId = container.get("containerId")
        self._domainName = container.get("domainName", [""])
        self._notes = container.get("notes", "")
        self._tagManagerUrl = container.get("tagManagerUrl")
        self._usageContext = container.get("usageContext")
        self._fingerprint = container.get("fingerprint")
        self._path = container.get("path")
        self._accountId = container.get("accountId")
        self._name = container.get("name")

        self._workspaces = []
        self._version_headers = []
        self._live_version = None
        self._raw_body = container

        self.workspaces = None

    def __repr__(self):
        return "{} {}".format(self.path, self.name)

    @property
    def publicId(self):
        """str: Container Public ID. Also known as the GTM-ID.
        """
        return self._publicId

    @property
    def containerId(self):
        """str: The Container ID uniquely identifies the GTM Container.
        """
        return self._containerId

    @property
    def domainName(self):
        """List[str]: List of domain names associated with the Container.
        """
        return self._domainName

    @property
    def notes(self):
        """str: Container Notes
        """
        return self._notes

    @property
    def tagManagerUrl(self):
        """str: Auto generated link to the tag manager UI
        """
        return self._tagManagerUrl

    @property
    def usageContext(self):
        """List[str]: List of Usage Contexts for the Container.
        Valid values include: web, android, or ios.
        """
        return self._usageContext

    @property
    def fingerprint(self):
        """str: The fingerprint of the GTM Container as computed
        at storage time. This value is recomputed whenever the account is modified.
        """
        return self._fingerprint

    @property
    def path(self):
        """str: GTM Container's API relative path.
        """
        return self._path

    @property
    def accountId(self):
        """str: GTM Account ID
        """
        return self._accountId

    @property
    def name(self):
        """str: Container display name
        """
        return self._name

    @property
    def raw_body(self):
        """obj: Container raw body
        """
        return self._raw_body

    def live_version(self, refresh=True):
        """Load from API and open the published GTM Version of the current container.

        Args:
            refresh (bool): If live_version has already been loaded from the API, force another API
                request to get the latest live_version.

        Returns:
            An instance of :class:`gtm_manager.version.GTMVersion` or `None` if the GTM Container
            has no published live version.

        """
        if refresh or not self._live_version:
            self._refresh_live_version()

        return self._live_version

    def create_workspace(self, name, description=""):
        """Create a new GTM Workspace in the current GTM Container.

        Args:
            name (str): Workspace display name
            description (str): Workspace description

        Returns:
            A new instance of the created :class:`gtm_manager.workspace.GTMWorkspace`.
        """
        body = {"description": description, "name": name}
        request = self.containers_service.workspaces().create(
            parent=self.path, body=body
        )
        response = request.execute()
        return gtm_manager.workspace.GTMWorkspace(
            workspace=response, service=self.service
        )

    def _get_container(self, path):
        """_get_container"""
        request = self.containers_service.get(path=path)
        response = request.execute()
        return response

    def _refresh_live_version(self):
        """_refresh_live_version"""
        try:
            request = self.containers_service.versions().live(parent=self.path)
            response = request.execute()
            self._live_version = gtm_manager.version.GTMVersion(
                version=response, service=self.service
            )
        except HttpError as error:
            if NO_LIVE_VERSION_ERROR in str(error):
                self._live_version = None
            else:
                raise error

    def list_workspaces(self, refresh=True):
        """Load from API and list all GTM Workspaces in this account.

        Args:
            refresh (bool): If workspaces have already been loaded from the API, force another API
                request to get the latest list of workspaces.

        Returns:
            A list of :class:`gtm_manager.workspace.GTMWorkspace`.
        """
        if self._workspaces is None or refresh:
            request = self.containers_service.workspaces().list(parent=self.path)
            response = request.execute()
            self._workspaces = [
                gtm_manager.workspace.GTMWorkspace(workspace=x, service=self.service)
                for x in response.get("workspace") or []
            ]
        return self._workspaces

    def list_version_headers(self, refresh=True):
        """Load from API and list all GTM Version Headers in this account.

        Args:
            refresh (bool): If version_headers have already been loaded from the API, force another
                API request to get the latest list of version_headers.

        Returns:
            A list of :class:`gtm_manager.version.GTMVersionHeader`.
        """
        if self._version_headers is None or refresh:
            request = (
                self.service.accounts()
                .containers()
                .version_headers()
                .list(parent=self.path)
            )
            response = request.execute()
            self._version_headers = [
                gtm_manager.version.GTMVersionHeader(versionHeader=x)
                for x in response.get("containerVersionHeader") or []
            ]
        return self._version_headers



"""version"""
import gtm_manager
import gtm_manager.container


class GTMVersion(gtm_manager.base.GTMBase):
    """Open a specific GTM Version.

    Args:
        version (dict): An API representation of the GTM Forlder. If provided, the resource will be
            not be loaded from the API. :code:`version` or :code:`path` argument must be set.
        path (str): The API path to the resource, i.e.
            "accounts/1234/containers/1234/version/123". If provided instead of
            :code:`version`, the representation will be loaded from the API. :code:`path` or
            :code:`version` argument must be set.
        parent (str): Required, when the instance is initialized with a :code:`version` argument to
            explizitly set the parent path, i.e. "accounts/1234/containers/1234/workspaces/1234"
        **kwargs: Additional keyword args to initialize the base class.
    """

    def __init__(self, version=None, path=None, workspaceId=None, **kwargs):
        super().__init__(**kwargs)

        self.versions_service = (
            self.service.accounts().containers().versions
        )  # pylint: disable=E1101

        if version:
            pass
        elif path:
            version = self._get_version(path)
        else:
            raise ValueError("Please pass either a version obj or version path.")

        self._container = gtm_manager.container.GTMContainer(
            container=version.get("container"), service=self.service
        )
        self._containerId = version.get("containerId")
        self._zone = version.get("zone")
        self._deleted = version.get("deleted")
        self._trigger = version.get("trigger") or []
        self._description = version.get("description")
        self._builtInVariable = version.get("builtInVariable")
        self._name = version.get("name")
        self._tag = version.get("tag") or []
        self._tagManagerUrl = version.get("tagManagerUrl")
        self._containerVersionId = version.get("containerVersionId")
        self._workspaceId = workspaceId
        self._fingerprint = version.get("fingerprint")
        self._variable = version.get("variable") or []
        self._path = version.get("path")
        self._folder = version.get("folder") or []
        self._accountId = version.get("accountId")

        workspace_path = (
            "accounts/{}/containers/{}/workspaces/{}".format(
                self._accountId, self._containerId, self._workspaceId
            )
            if workspaceId
            else None
        )

        self._tag = [
            gtm_manager.tag.GTMTag(tag=x, parent=workspace_path, service=self.service)
            for x in self._tag
        ]
        self._trigger = [
            gtm_manager.trigger.GTMTrigger(
                trigger=x, parent=workspace_path, service=self.service
            )
            for x in self._trigger
        ]
        self._variable = [
            gtm_manager.variable.GTMVariable(
                variable=x, parent=workspace_path, service=self.service
            )
            for x in self._variable
        ]
        self._folder = [
            gtm_manager.folder.GTMFolder(
                folder=x, parent=workspace_path, service=self.service
            )
            for x in self._folder
        ]

        self.raw_body = version

    @property
    def container(self):
        """:class:`gtm_manager.container.GTMContainer`: The container that this version was taken
        from.
        """
        return self._container

    @property
    def containerId(self):
        """str: GTM Container ID.
        """
        return self._containerId

    @property
    def zone(self):
        """list: The zones in the container that this version was taken from.
        """
        return self._zone

    @property
    def deleted(self):
        """bool: A value of true indicates this container version has been deleted.
        """
        return self._deleted

    @property
    def trigger(self):
        """list: The triggers in the container that this version was taken from.
        """
        return self._trigger

    @property
    def description(self):
        """str: Container version description.
        """
        return self._description

    @property
    def builtInVariable(self):
        """list: The built-in variables in the container that this version was taken from.
        """
        return self._builtInVariable

    @property
    def name(self):
        """str: Container version display name.
        """
        return self._name

    @property
    def tag(self):
        """list: The tags in the container that this version was taken from.
        """
        return self._tag

    @property
    def tagManagerUrl(self):
        """str: Auto generated link to the tag manager UI
        """
        return self._tagManagerUrl

    @property
    def containerVersionId(self):
        """str: The Container Version ID uniquely identifies the GTM Container Version.
        """
        return self._containerVersionId

    @property
    def fingerprint(self):
        """str: The fingerprint of the GTM Container Version as computed at storage
        time. This value is recomputed whenever the container version is modified.
        """
        return self._fingerprint

    @property
    def variable(self):
        """list: The variables in the container that this version was taken from.
        """
        return self._variable

    @property
    def path(self):
        """str: GTM ContainerVersions's API relative path.
        """
        return self._path

    @property
    def folder(self):
        """list: The folders in the container that this version was taken from.
        """
        return self._folder

    @property
    def accountId(self):
        """str: GTM Account ID.
        """
        return self._accountId

    def _get_version(self, path):
        """_get_version"""
        request = self.versions_service().get(path=path)
        response = request.execute()
        return response

    def publish(self):
        """publish"""
        request = self.versions_service().publish(path=self.path)
        response = request.execute()
        return response


class GTMVersionHeader(object):
    """GTMVersionHeader"""

    def __init__(self, versionHeader):
        self._path = versionHeader.get("path")
        self._accountId = versionHeader.get("accountId")
        self._containerId = versionHeader.get("containerId")
        self._containerVersionId = versionHeader.get("containerVersionId")
        self._name = versionHeader.get("name")
        self._numMacros = versionHeader.get("numMacros")
        self._numRules = versionHeader.get("numRules")
        self._numTags = versionHeader.get("numTags")
        self._numTriggers = versionHeader.get("numTriggers")
        self._deleted = versionHeader.get("deleted")
        self._numVariables = versionHeader.get("numVariables")
        self._numZones = versionHeader.get("numZones")

        self.raw_body = versionHeader
